"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2021 Deepkit UG, Marc J. Schmidt
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the MIT License.
 *
 * You should have received a copy of the MIT License along with this program.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PostgresDatabaseAdapter = exports.PostgresSQLDatabaseQueryFactory = exports.PostgresSQLDatabaseQuery = exports.PostgresSQLQueryResolver = exports.PostgresPersistence = exports.PostgresConnectionPool = exports.PostgresDatabaseTransaction = exports.PostgresConnection = exports.PostgresStatement = void 0;
const sql_1 = require("@deepkit/sql");
const orm_1 = require("@deepkit/orm");
const postgres_platform_1 = require("./postgres-platform");
const type_1 = require("@deepkit/type");
const pg_1 = __importDefault(require("pg"));
const core_1 = require("@deepkit/core");
const stopwatch_1 = require("@deepkit/stopwatch");
function handleError(error) {
    const message = 'string' === typeof error ? error : error.message;
    if (message.includes('violates unique constraint')) {
        //todo: extract table name, column name, and find ClassSchema
        throw new orm_1.UniqueConstraintFailure();
    }
}
class PostgresStatement extends sql_1.SQLStatement {
    constructor(logger, sql, client, stopwatch) {
        super();
        this.logger = logger;
        this.sql = sql;
        this.client = client;
        this.stopwatch = stopwatch;
        this.released = false;
    }
    async get(params = []) {
        const frame = this.stopwatch ? this.stopwatch.start('Query', stopwatch_1.FrameCategory.databaseQuery) : undefined;
        try {
            if (frame)
                frame.data({ sql: this.sql, sqlParams: params });
            //postgres driver does not maintain error.stack when they throw errors, so
            //we have to manually convert it using asyncOperation.
            const res = await core_1.asyncOperation((resolve, reject) => {
                this.client.query(this.sql, params).then(resolve).catch(reject);
            });
            return res.rows[0];
        }
        catch (error) {
            handleError(error);
            this.logger.failedQuery(error, this.sql, params);
            throw error;
        }
        finally {
            if (frame)
                frame.end();
        }
    }
    async all(params = []) {
        const frame = this.stopwatch ? this.stopwatch.start('Query', stopwatch_1.FrameCategory.databaseQuery) : undefined;
        try {
            if (frame)
                frame.data({ sql: this.sql, sqlParams: params });
            //postgres driver does not maintain error.stack when they throw errors, so
            //we have to manually convert it using asyncOperation.
            const res = await core_1.asyncOperation((resolve, reject) => {
                this.client.query(this.sql, params).then(resolve).catch(reject);
            });
            return res.rows;
        }
        catch (error) {
            handleError(error);
            this.logger.failedQuery(error, this.sql, params);
            throw error;
        }
        finally {
            if (frame)
                frame.end();
        }
    }
    release() {
    }
}
exports.PostgresStatement = PostgresStatement;
class PostgresConnection extends sql_1.SQLConnection {
    constructor(connectionPool, connection, logger, transaction, stopwatch) {
        super(connectionPool, logger, transaction, stopwatch);
        this.connection = connection;
        this.changes = 0;
        this.lastReturningRows = [];
    }
    async prepare(sql) {
        return new PostgresStatement(this.logger, sql, this.connection, this.stopwatch);
    }
    async run(sql, params = []) {
        const frame = this.stopwatch ? this.stopwatch.start('Query', stopwatch_1.FrameCategory.databaseQuery) : undefined;
        try {
            if (frame)
                frame.data({ sql, sqlParams: params });
            //postgres driver does not maintain error.stack when they throw errors, so
            //we have to manually convert it using asyncOperation.
            const res = await core_1.asyncOperation((resolve, reject) => {
                this.connection.query(sql, params).then(resolve).catch(reject);
            });
            this.logger.logQuery(sql, params);
            this.lastReturningRows = res.rows;
            this.changes = res.rowCount;
        }
        catch (error) {
            handleError(error);
            this.logger.failedQuery(error, sql, params);
            throw error;
        }
        finally {
            if (frame)
                frame.end();
        }
    }
    async getChanges() {
        return this.changes;
    }
}
exports.PostgresConnection = PostgresConnection;
class PostgresDatabaseTransaction extends orm_1.DatabaseTransaction {
    /**
     * This is the default for mysql databases.
     */
    repeatableRead() {
        this.setTransaction = 'REPEATABLE READ';
        return this;
    }
    readCommitted() {
        this.setTransaction = 'READ COMMITTED';
        return this;
    }
    serializable() {
        this.setTransaction = 'SERIALIZABLE';
        return this;
    }
    async begin() {
        if (!this.connection)
            return;
        const set = this.setTransaction ? 'SET TRANSACTION ISOLATION LEVEL ' + this.setTransaction + ';' : '';
        await this.connection.run(set + 'START TRANSACTION');
    }
    async commit() {
        if (!this.connection)
            return;
        if (this.ended)
            throw new Error('Transaction ended already');
        await this.connection.run('COMMIT');
        this.ended = true;
        this.connection.release();
    }
    async rollback() {
        if (!this.connection)
            return;
        if (this.ended)
            throw new Error('Transaction ended already');
        await this.connection.run('ROLLBACK');
        this.ended = true;
        this.connection.release();
    }
}
exports.PostgresDatabaseTransaction = PostgresDatabaseTransaction;
class PostgresConnectionPool extends sql_1.SQLConnectionPool {
    constructor(pool) {
        super();
        this.pool = pool;
    }
    async getConnection(logger, transaction, stopwatch) {
        //when a transaction object is given, it means we make the connection sticky exclusively to that transaction
        //and only release the connection when the transaction is commit/rollback is executed.
        if (transaction && transaction.connection)
            return transaction.connection;
        const poolClient = await this.pool.connect();
        this.activeConnections++;
        const connection = new PostgresConnection(this, poolClient, logger, transaction, stopwatch);
        if (transaction) {
            transaction.connection = connection;
            try {
                await transaction.begin();
            }
            catch (error) {
                transaction.ended = true;
                connection.release();
                throw new Error('Could not start transaction: ' + error);
            }
        }
        return connection;
    }
    release(connection) {
        //connections attached to a transaction are not automatically released.
        //only with commit/rollback actions
        if (connection.transaction && !connection.transaction.ended)
            return;
        super.release(connection);
        connection.connection.release();
    }
}
exports.PostgresConnectionPool = PostgresConnectionPool;
function typeSafeDefaultValue(property) {
    if (property.type === 'string')
        return '';
    if (property.type === 'number')
        return 0;
    if (property.type === 'boolean')
        return false;
    if (property.type === 'date')
        return new Date;
    return null;
}
class PostgresPersistence extends sql_1.SQLPersistence {
    constructor(platform, connectionPool, session) {
        super(platform, connectionPool, session);
        this.platform = platform;
        this.connectionPool = connectionPool;
        this.placeholderPosition = 1;
    }
    async batchUpdate(classSchema, changeSets) {
        const scopeSerializer = this.platform.serializer.for(classSchema);
        const tableName = this.platform.getTableIdentifier(classSchema);
        const pkName = classSchema.getPrimaryField().name;
        const pkField = this.platform.quoteIdentifier(pkName);
        const values = {};
        const setNames = [];
        const aggregateSelects = {};
        const requiredFields = {};
        const assignReturning = {};
        const setReturning = {};
        for (const changeSet of changeSets) {
            const where = [];
            const pk = scopeSerializer.partialSerialize(changeSet.primaryKey);
            for (const i in pk) {
                if (!pk.hasOwnProperty(i))
                    continue;
                where.push(`${this.platform.quoteIdentifier(i)} = ${this.platform.quoteValue(pk[i])}`);
                requiredFields[i] = 1;
            }
            if (!values[pkName])
                values[pkName] = [];
            values[pkName].push(this.platform.quoteValue(changeSet.primaryKey[pkName]));
            const fieldAddedToValues = {};
            const id = changeSet.primaryKey[pkName];
            if (changeSet.changes.$set) {
                const value = scopeSerializer.partialSerialize(changeSet.changes.$set);
                for (const i in value) {
                    if (!value.hasOwnProperty(i))
                        continue;
                    if (!values[i]) {
                        values[i] = [];
                        setNames.push(`${this.platform.quoteIdentifier(i)} = _b.${this.platform.quoteIdentifier(i)}`);
                    }
                    requiredFields[i] = 1;
                    fieldAddedToValues[i] = 1;
                    let v = value[i];
                    //special postgres check to avoid an error like:
                    /// column "deletedAt" is of type timestamp without time zone but expression is of type text
                    if (v === undefined || v === null) {
                        values[i].push('null' + this.platform.typeCast(classSchema, i));
                    }
                    else {
                        values[i].push(this.platform.quoteValue(v));
                    }
                }
            }
            if (changeSet.changes.$inc) {
                for (const i in changeSet.changes.$inc) {
                    if (!changeSet.changes.$inc.hasOwnProperty(i))
                        continue;
                    const value = changeSet.changes.$inc[i];
                    if (!aggregateSelects[i])
                        aggregateSelects[i] = [];
                    if (!values[i]) {
                        values[i] = [];
                        setNames.push(`${this.platform.quoteIdentifier(i)} = _b.${this.platform.quoteIdentifier(i)}`);
                    }
                    if (!assignReturning[id]) {
                        assignReturning[id] = { item: changeSet.item, names: [] };
                    }
                    assignReturning[id].names.push(i);
                    setReturning[i] = 1;
                    aggregateSelects[i].push({
                        id: changeSet.primaryKey[pkName],
                        sql: `_origin.${this.platform.quoteIdentifier(i)} + ${this.platform.quoteValue(value)}`
                    });
                    requiredFields[i] = 1;
                    if (!fieldAddedToValues[i]) {
                        fieldAddedToValues[i] = 1;
                        values[i].push(this.platform.quoteValue(typeSafeDefaultValue(classSchema.getProperty(i))));
                    }
                }
            }
        }
        const selects = [];
        const valuesValues = [];
        const valuesNames = [];
        for (const i in values) {
            valuesNames.push(i);
        }
        for (let i = 0; i < values[pkName].length; i++) {
            valuesValues.push('(' + valuesNames.map(name => values[name][i]).join(',') + ')');
        }
        for (const i in requiredFields) {
            if (aggregateSelects[i]) {
                const select = [];
                select.push('CASE');
                for (const item of aggregateSelects[i]) {
                    select.push(`WHEN _.${pkField} = ${item.id} THEN ${item.sql}`);
                }
                select.push(`ELSE _.${this.platform.quoteIdentifier(i)} END as ${this.platform.quoteIdentifier(i)}`);
                selects.push(select.join(' '));
            }
            else {
                selects.push('_.' + this.platform.quoteIdentifier(i));
            }
        }
        const returningSelect = [];
        returningSelect.push(tableName + '.' + pkField);
        if (!core_1.empty(setReturning)) {
            for (const i in setReturning) {
                returningSelect.push(tableName + '.' + this.platform.quoteIdentifier(i));
            }
        }
        const escapedValuesNames = valuesNames.map(v => this.platform.quoteIdentifier(v));
        const sql = `
              WITH _b(${escapedValuesNames.join(', ')}) AS (
                SELECT ${selects.join(', ')} FROM
                    (VALUES ${valuesValues.join(', ')}) as _(${escapedValuesNames.join(', ')})
                    INNER JOIN ${tableName} as _origin ON (_origin.${pkField} = _.${pkField})
              )
              UPDATE ${tableName}
              SET ${setNames.join(', ')}
              FROM _b
              WHERE ${tableName}.${pkField} = _b.${pkField}
              RETURNING ${returningSelect.join(', ')};
        `;
        const connection = await this.getConnection(); //will automatically be released in SQLPersistence
        const result = await connection.execAndReturnAll(sql);
        for (const returning of result) {
            const r = assignReturning[returning[pkName]];
            if (!r)
                continue;
            for (const name of r.names) {
                r.item[name] = returning[name];
            }
        }
    }
    async populateAutoIncrementFields(classSchema, items) {
        const autoIncrement = classSchema.getAutoIncrementField();
        if (!autoIncrement)
            return;
        const connection = await this.getConnection(); //will automatically be released in SQLPersistence
        //We adjusted the INSERT SQL with additional RETURNING which returns all generated
        //auto-increment values. We read the result and simply assign the value.
        const name = autoIncrement.name;
        const insertedRows = connection.lastReturningRows;
        if (!insertedRows.length)
            return;
        for (let i = 0; i < items.length; i++) {
            items[i][name] = insertedRows[i][name];
        }
    }
    getInsertSQL(classSchema, fields, values) {
        const autoIncrement = classSchema.getAutoIncrementField();
        const returning = autoIncrement ? ` RETURNING ${this.platform.quoteIdentifier(autoIncrement.name)}` : '';
        return `INSERT INTO ${this.platform.getTableIdentifier(classSchema)} (${fields.join(', ')}) VALUES (${values.join('), (')}) ${returning}`;
    }
    resetPlaceholderSymbol() {
        this.placeholderPosition = 1;
    }
    getPlaceholderSymbol() {
        return '$' + this.placeholderPosition++;
    }
}
exports.PostgresPersistence = PostgresPersistence;
class PostgresSQLQueryResolver extends sql_1.SQLQueryResolver {
    async delete(model, deleteResult) {
        const primaryKey = this.classSchema.getPrimaryField();
        const pkField = this.platform.quoteIdentifier(primaryKey.name);
        const primaryKeyConverted = type_1.getPropertyXtoClassFunction(primaryKey, this.platform.serializer);
        const sqlBuilder = new sql_1.SqlBuilder(this.platform);
        const select = sqlBuilder.select(this.classSchema, model, { select: [pkField] });
        const tableName = this.platform.getTableIdentifier(this.classSchema);
        const connection = await this.connectionPool.getConnection(this.session.logger, this.session.assignedTransaction, this.session.stopwatch);
        try {
            const sql = `
                WITH _ AS (${select.sql})
                DELETE
                FROM ${tableName} USING _
                WHERE ${tableName}.${pkField} = _.${pkField}
                RETURNING ${tableName}.${pkField}
            `;
            const rows = await connection.execAndReturnAll(sql, select.params);
            deleteResult.modified = rows.length;
            for (const row of rows) {
                deleteResult.primaryKeys.push(primaryKeyConverted(row[primaryKey.name]));
            }
        }
        finally {
            connection.release();
        }
    }
    async patch(model, changes, patchResult) {
        const select = [];
        const selectParams = [];
        const tableName = this.platform.getTableIdentifier(this.classSchema);
        const primaryKey = this.classSchema.getPrimaryField();
        const primaryKeyConverted = type_1.getPropertyXtoClassFunction(primaryKey, this.platform.serializer);
        const fieldsSet = {};
        const aggregateFields = {};
        const scopeSerializer = this.platform.serializer.for(this.classSchema);
        const $set = changes.$set ? scopeSerializer.partialSerialize(changes.$set) : undefined;
        const set = [];
        if ($set)
            for (const i in $set) {
                if (!$set.hasOwnProperty(i))
                    continue;
                if ($set[i] === undefined || $set[i] === null) {
                    set.push(`${this.platform.quoteIdentifier(i)} = NULL`);
                }
                else {
                    fieldsSet[i] = 1;
                    select.push(`$${selectParams.length + 1}${this.platform.typeCast(this.classSchema, i)} as ${this.platform.quoteIdentifier(i)}`);
                    selectParams.push($set[i]);
                }
            }
        if (changes.$unset)
            for (const i in changes.$unset) {
                if (!changes.$unset.hasOwnProperty(i))
                    continue;
                fieldsSet[i] = 1;
                select.push(`NULL as ${this.platform.quoteIdentifier(i)}`);
            }
        for (const i of model.returning) {
            aggregateFields[i] = { converted: type_1.getPropertyXtoClassFunction(type_1.resolvePropertySchema(this.classSchema, i), this.platform.serializer) };
            select.push(`(${this.platform.quoteIdentifier(i)} ) as ${this.platform.quoteIdentifier(i)}`);
        }
        if (changes.$inc)
            for (const i in changes.$inc) {
                if (!changes.$inc.hasOwnProperty(i))
                    continue;
                fieldsSet[i] = 1;
                aggregateFields[i] = { converted: type_1.getPropertyXtoClassFunction(type_1.resolvePropertySchema(this.classSchema, i), this.platform.serializer) };
                select.push(`(${this.platform.quoteIdentifier(i)} + ${this.platform.quoteValue(changes.$inc[i])}) as ${this.platform.quoteIdentifier(i)}`);
            }
        for (const i in fieldsSet) {
            set.push(`${this.platform.quoteIdentifier(i)} = _b.${this.platform.quoteIdentifier(i)}`);
        }
        let bPrimaryKey = primaryKey.name;
        //we need a different name because primaryKeys could be updated as well
        if (fieldsSet[primaryKey.name]) {
            select.unshift(this.platform.quoteIdentifier(primaryKey.name) + ' as __' + primaryKey.name);
            bPrimaryKey = '__' + primaryKey.name;
        }
        else {
            select.unshift(this.platform.quoteIdentifier(primaryKey.name));
        }
        const returningSelect = [];
        returningSelect.push(tableName + '.' + this.platform.quoteIdentifier(primaryKey.name));
        if (!core_1.empty(aggregateFields)) {
            for (const i in aggregateFields) {
                returningSelect.push(tableName + '.' + this.platform.quoteIdentifier(i));
            }
        }
        const sqlBuilder = new sql_1.SqlBuilder(this.platform, selectParams);
        const selectSQL = sqlBuilder.select(this.classSchema, model, { select });
        const sql = `
            WITH _b AS (${selectSQL.sql})
            UPDATE
                ${tableName}
            SET
                ${set.join(', ')}
            FROM _b
            WHERE ${tableName}.${this.platform.quoteIdentifier(primaryKey.name)} = _b.${this.platform.quoteIdentifier(bPrimaryKey)}
            RETURNING ${returningSelect.join(', ')}
        `;
        const connection = await this.connectionPool.getConnection(this.session.logger, this.session.assignedTransaction, this.session.stopwatch);
        try {
            const result = await connection.execAndReturnAll(sql, selectSQL.params);
            patchResult.modified = result.length;
            for (const i in aggregateFields) {
                patchResult.returning[i] = [];
            }
            for (const returning of result) {
                patchResult.primaryKeys.push(primaryKeyConverted(returning[primaryKey.name]));
                for (const i in aggregateFields) {
                    patchResult.returning[i].push(aggregateFields[i].converted(returning[i]));
                }
            }
        }
        finally {
            connection.release();
        }
    }
}
exports.PostgresSQLQueryResolver = PostgresSQLQueryResolver;
class PostgresSQLDatabaseQuery extends sql_1.SQLDatabaseQuery {
}
exports.PostgresSQLDatabaseQuery = PostgresSQLDatabaseQuery;
class PostgresSQLDatabaseQueryFactory extends sql_1.SQLDatabaseQueryFactory {
    createQuery(classType) {
        return new PostgresSQLDatabaseQuery(type_1.getClassSchema(classType), this.databaseSession, new PostgresSQLQueryResolver(this.connectionPool, this.platform, type_1.getClassSchema(classType), this.databaseSession));
    }
}
exports.PostgresSQLDatabaseQueryFactory = PostgresSQLDatabaseQueryFactory;
class PostgresDatabaseAdapter extends sql_1.SQLDatabaseAdapter {
    constructor(options) {
        super();
        this.options = options;
        this.pool = new pg_1.default.Pool(this.options);
        this.connectionPool = new PostgresConnectionPool(this.pool);
        this.platform = new postgres_platform_1.PostgresPlatform();
        pg_1.default.types.setTypeParser(1700, parseFloat);
        pg_1.default.types.setTypeParser(20, parseInt);
    }
    getName() {
        return 'postgres';
    }
    getSchemaName() {
        //todo extract schema name from connection options. This acts as default when a table has no schemaName defined.
        return '';
    }
    createPersistence(session) {
        return new PostgresPersistence(this.platform, this.connectionPool, session);
    }
    createTransaction(session) {
        return new PostgresDatabaseTransaction;
    }
    queryFactory(session) {
        return new PostgresSQLDatabaseQueryFactory(this.connectionPool, this.platform, session);
    }
    disconnect(force) {
        this.pool.end().catch(console.error);
    }
}
exports.PostgresDatabaseAdapter = PostgresDatabaseAdapter;
//# sourceMappingURL=postgres-adapter.js.map